#!/usr/bin/env python3
# coding: utf-8

"""Loop simulator

Simulate Hi-C datasets with chromatin loops, with parameters based on
experimental data, for the purpose of benchmarking, training, and testing.

"""


def main():
    pass


if __name__ == "__main__":
    main()
